"""
OPFONMW = Other Project From Other, Not Me Write
use for easy code

dearpygui_animate = mrtnRitter (https://github.com/mrtnRitter/)
ofdm_codec = berndporr (https://github.com/berndporr/py_ofdm)
StepperLib = benkjor (https://github.com/benkjor)
"""